#!/opt/ias/env/bin/python
import argparse

import sh


# 参考: http://amoffat.github.io/sh/


def main():
    parser = argparse.ArgumentParser("ffmpeg文件分割工具")  # TODO
    parser.add_argument("src_dir", type=Path, help="文件来源目录")
    parser.add_argument("-e", "--ext", type=str, default="jpg", help="文件扩展名")
    parser.add_argument(
        "-m", "--mate_ext", type=str, default="lbl", help="文件伙伴文件扩展名"
    )
    parser.add_argument(
        "-i", "--invert_match", action="store_true", help="选中不匹配的文件"
    )
    parser.add_argument("-v", "--verbose", action="store_true", help="显示详细信息")
    opt = parser.parse_args()

    mates = files_in(opt.src_dir, opt.mate_ext)
    names = [f.stem for f in mates]


if __name__ == "__main__":
    main()



from typing import Optional, TypeAlias

import cv2
import numpy as np
from jcx.sys.fs import StrPath
from jvi.image.image_nda import ImageNda


def map_color(
    src: ImageNda, colormap: int | np.ndarray, dst: Optional[ImageNda] = None
) -> ImageNda:
    """灰度图转换为彩色图"""
    # 参考: https://docs.opencv.org/4.x/d3/d50/group__imgproc__colormap.html
    # assert src.channel_num() == 1, '输入图像必须为灰度图'
    if dst:
        assert dst.same_size(src), "输出图片必须与输入图片尺寸相同"
    dd = cv2.applyColorMap(src.data(), colormap, dst and dst.data())
    assert isinstance(dd, np.ndarray)
    return ImageNda(data=dd)


PaletteHS: TypeAlias = list[tuple[int, int]]
"""调色板类型定义"""


def load_palette_hs(path: StrPath) -> PaletteHS:
    """从调色板图片加载(色度,饱和度)调色板数据"""
    im = ImageNda.load(path)
    mat = im.data()
    bgr_col = np.ascontiguousarray(mat[:, 1:2])  # 只截取一列

    hsv_col = cv2.cvtColor(bgr_col, cv2.COLOR_BGR2HLS_FULL)
    hs_col = hsv_col[:, :, 0:2]  # 去掉亮度通道
    # print('h1 max:', hs_col.max())

    palette = hs_col.reshape(-1, 2)  # 变为(h, s)对数组, 温度:高->低
    palette = palette[::-1]  # 逆序, 温度:低->高
    return [(h, s) for h, s in palette.tolist()]


def extract_image_ir(
    palette: PaletteHS, im_src: ImageNda, im_dst: Optional[ImageNda] = None
) -> ImageNda:
    """利用调色版从伪彩色热图中分类单通道热图"""
    im_dst = im_dst or ImageNda.new_as(im_src, channel=1)  # 同尺寸灰度图

    assert im_src.same_size(im_dst), "图像尺寸不一致"
    assert im_src.channel_num() == 3, "输入图像必须为BGR格式"
    assert im_dst.channel_num() == 1, "输入图像必须为单通"

    mat_hsv = cv2.cvtColor(im_src.data(), cv2.COLOR_BGR2HLS_FULL)  # 转成HSV图
    mat_hs = mat_hsv[:, :, 0:2]  # 只取H和S通道
    arr_hs = mat_hs.reshape(-1, 2)  # 转成点数组
    mat_dst = im_dst.data().reshape(-1)  # 用一维方式访问图像
    for i, (h, s) in enumerate(arr_hs):  # 遍历HS图像
        # print(f'#{i} {h} {s}')
        d2_min = 100000
        idx_min = 0
        for j in range(len(palette)):  # 搜索最接近的调色板项
            h1, s1 = palette[j]
            d2 = (h - h1) ** 2 + (s - s1) ** 2
            if d2 < d2_min:
                d2_min = d2
                idx_min = j
        mat_dst[i] = idx_min
    return im_dst


def to_color(src: ImageNda, dst: Optional[ImageNda] = None) -> ImageNda:
    """灰度图转换为彩色图"""
    """
    
    rainbow_file = JIV_STATIC / 'infrared/rainbow.jpg'
    rainbow = ImageNda.load(rainbow_file)
    m = rainbow.data()
    rainbow_color = np.ascontiguousarray(m[:, 1:2])
    print(m.shape, rainbow_color.shape)

    src_file = JIV_STATIC / 'infrared/wh1.jpg'
    src = ImageNda.load(src_file)
    src_1c = to_gray(src)
    dst = map_color(src_1c, rainbow_color) # 彩虹调色板

    print(src, dst)

    trace_images([src, dst], 'Infrared', box_size=Size(1920, 540))
    close_all_windows()

    """
    return map_color(src, cv2.COLORMAP_JET, dst)



import re
from copy import copy
from typing import Optional, Final, Self

from jcx.m.number import real_2d, Real1_2D, Real
from pydantic import BaseModel

from jvi.geo import is_normalized
from jvi.geo.point2d import Points, Point


class Size(BaseModel):
    """二维尺寸"""

    width: float = 0
    height: float = 0

    @classmethod
    def new(cls, width: float, height: float) -> Self:
        """长方形-构造函数"""
        return Size(width=width, height=height)

    @classmethod
    def zero(cls) -> Self:
        """长方形-构造函数"""
        return Size()

    def size(self) -> Self:
        """实现ISize2D"""
        return self

    def __bool__(self) -> bool:
        """判定是否存在非零属性"""
        return self != Size()

    def brief(self) -> str:
        """获取简要描述"""
        return "{}x{}".format(self.width, self.height)

    def area(self) -> Real:
        """获取面积"""
        return self.width * self.height

    def empty(self) -> bool:
        """判定面积为零"""
        return self.area() == 0

    def positive(self) -> bool:
        """判定长宽为正"""
        return self.width > 0 and self.height > 0

    def center(self) -> Point:
        """获取中心坐标"""
        return Point(x=self.width / 2, y=self.height / 2)

    def aspect_ratio(self) -> float:
        """获取宽高比"""
        return self.width / self.height

    def contains(self, other: Self) -> bool:
        """是否包含目标尺寸"""
        return self.width >= other.width and self.height >= other.height

    def clone(self) -> Self:
        """克隆对象"""
        return copy(self)

    def set(self, other: Self) -> None:
        """设置为目标对象值"""
        self.width = other.width
        self.height = other.height

    def scale_me(self, n: Real1_2D) -> None:
        """缩放指定的倍数"""
        x, y = real_2d(n)
        self.width *= x
        self.height *= y

    def scale(self, n: Real1_2D) -> Self:
        """缩放指定的倍数"""
        s = self.clone()
        s.scale_me(n)
        return s

    def to_tuple(self) -> tuple[float, float]:
        """获取tuple"""
        return self.width, self.height

    def to_tuple_int(self) -> tuple[int, int]:
        """转换为tuple"""
        return int(self.width), int(self.height)

    def to_shape(self) -> tuple[int, int]:
        """转换为 2D Shape"""
        return int(self.height), int(self.width)

    def to_shape3d_i(self, channel: int) -> tuple[int, int, int]:
        """转换为 3D Shape交错格式"""
        return int(self.height), int(self.width), channel

    def to_shape3d_p(self, channel: int) -> tuple[int, int, int]:
        """转换为 3D Shape平坦格式"""
        return channel, int(self.height), int(self.width)

    def to_point(self) -> Point:
        """转换为Point"""
        return Point(x=self.width, y=self.height)

    def normalize_me(self, size: Self) -> None:
        """绝对坐标归一化"""
        self.width /= size.width
        self.height /= size.height

    def normalize(self, size: Self) -> Self:
        """获取绝对坐标归一化"""
        p = copy(self)
        p.normalize_me(size)
        return p

    def absolutize_me(self, size: Self) -> None:
        """归一化坐标绝对化"""
        self.width *= size.width
        self.height *= size.height
        self.round_me()  # 存在非整数可能

    def absolutize(self, size: Self) -> Self:
        """获取归一化坐标绝对化"""
        p = copy(self)
        p.absolutize_me(size)
        return p

    def round_me(self) -> None:
        """近似成整数"""
        self.width = round(self.width)
        self.height = round(self.height)

    def round(self) -> Self:
        """近似成整数"""
        s = copy(self)
        s.round_me()
        return s

    def is_normalized(self) -> bool:
        """判断坐标是否被归一化"""
        return is_normalized(self.width) and is_normalized(self.height)

    def __mul__(self, s: Self) -> Self:
        """尺寸相乘"""
        return Size(width=self.width * s.width, height=self.height * s.height)

    def __truediv__(self, s: Self) -> Self:
        """尺寸相除"""
        return Size(width=self.width / s.width, height=self.height / s.height)

    def __or__(self, s: Self) -> Self:
        """尺寸并集"""
        return Size(width=max(self.width, s.width), height=max(self.height, s.height))

    def __and__(self, s: Self) -> Self:
        """尺寸并集"""
        return Size(width=min(self.width, s.width), height=min(self.height, s.height))

    def scale_in(self, box_size: Self) -> Self:
        """保证宽高比缩放到指定尺寸以内"""
        rx = box_size.width / self.width
        ry = box_size.height / self.height
        r = min(rx, ry)

        w = self.width * r
        h = self.height * r
        return Size(width=w, height=h)


def absolutize_points(points: Points, size: Size) -> Points:
    """点集归一化坐标绝对化"""
    arr = [p.absolutize(size) for p in points]
    return arr


# 参考：https://zh.wikipedia.org/wiki/%E6%98%BE%E7%A4%BA%E5%88%86%E8%BE%A8%E7%8E%87%E5%88%97%E8%A1%A8

SIZE_8K: Final = Size(width=8192, height=4320)
SIZE_8K_UHD: Final = Size(width=7680, height=4320)
SIZE_DCI_4K: Final = Size(width=4096, height=2160)
SIZE_4K_UHD: Final = Size(width=3840, height=2160)
SIZE_3K: Final = Size(width=2880, height=1620)
SIZE_2K: Final = Size(width=2048, height=1080)
SIZE_FHD: Final = Size(width=1920, height=1080)
SIZE_qHD: Final = Size(width=960, height=540)
SIZE_HD_PLUS: Final = Size(width=1600, height=900)
SIZE_HD: Final = Size(width=1280, height=720)
SIZE_nHD: Final = Size(width=640, height=360)
SIZE_VGA: Final = Size(width=640, height=480)
SIZE_QVGA: Final = Size(width=320, height=240)
SIZE_PAL: Final = Size(width=768, height=576)
SIZE_IM: Final = Size(width=224, height=224)

SIZE_640: Final = Size(width=640, height=640)
"""深度学习检测器常规尺寸"""
SIZE_720: Final = Size(width=720, height=720)
"""深度学习检测器常规尺寸"""
SIZE_1280: Final = Size(width=1280, height=1280)
"""深度学习检测器常规尺寸"""

resolutions: Final = {
    "8K": SIZE_8K,
    "8K_UHD": SIZE_8K_UHD,
    "DCI_4K": SIZE_DCI_4K,
    "4K_UHD": SIZE_4K_UHD,
    "3K": SIZE_3K,
    "2K": SIZE_2K,
    "FHD": SIZE_FHD,
    "qHD": SIZE_qHD,
    "HD+": SIZE_HD_PLUS,
    "HD": SIZE_HD,
    "nHD": SIZE_nHD,
    "VGA": SIZE_VGA,
    "QVGA": SIZE_QVGA,
    "PAL": SIZE_PAL,
    "IM": SIZE_IM,
}


def size_parse(s: str) -> Optional[Size]:
    """解析字符串获取尺寸"""

    size = resolutions.get(s)
    if size:
        return size

    p = re.compile(r"(\d+)x(\d+)")
    m = p.match(s.lower())
    if m:
        w = int(m.groups()[0])
        h = int(m.groups()[1])
        size = Size(width=w, height=h)
    else:
        size = None
    return size



#!/opt/ias/env/bin/python

import argparse
from pathlib import Path

import cv2  # type: ignore


def main():
    parser = argparse.ArgumentParser(description="图片模糊工具")
    parser.add_argument("src_dir", type=Path, help="来源图片目录")
    parser.add_argument("dst_dir", type=Path, help="目的图片目录")
    parser.add_argument("level", type=int, help="视频模糊级别")
    parser.add_argument(
        "-p", "--pattern", type=str, default="*.jpg", help="图片文件模式(*.jpg)"
    )
    parser.add_argument("-t", "--tail", type=str, default="_b", help="图片后缀")
    parser.add_argument("-v", "--verbose", action="store_true", help="显示详细信息")
    opt = parser.parse_args()

    files = opt.src_dir.rglob(opt.pattern)
    print("源图片目录：%s" % opt.src_dir)

    n = 2 * opt.level + 1
    ksize = (n, n)

    for i, f in enumerate(files):
        src = ImageNda.load(str(f))  # BGR
        if src is None:
            print(i, f.name, "\tERROR")
            continue
        if opt.verbose:
            print(i, f.name, "\tOK")
        dst = cv2.blur(src, ksize)

        name = f.stem + opt.tail + str(opt.level) + f.suffix
        file = opt.dst_dir / name
        print(file)
        cv2.imwrite(str(file), dst)
        # cv2.imshow('src', src)
        # cv2.imshow('dst', dst)
        # cv2.waitKey(0)

    # cv2.destroyAllWindows()


if __name__ == "__main__":
    main()



from math import isclose

import cv2
import numpy as np

from jvi.drawing.color import Color
from jvi.image.image_nda import ImageNda


def value_map(img_1c: ImageNda) -> dict[int, float]:
    """获取图像分布"""
    assert img_1c.channel_num() == 1
    # https://numpy.org/doc/stable/reference/generated/numpy.histogram.html
    # 参见：[Python numpy.histogram函数方法的使用](https://www.cjavapy.com/article/1103/)
    hist, edge = np.histogram(img_1c.data(), 256, (0, 255), density=True)
    print("hist1.shape:", hist.shape, edge.shape)

    m = {i: v for i, v in enumerate(hist) if v > 0}
    return m


def hue_hist(image: ImageNda, bins: int = 256) -> tuple[np.ndarray, np.ndarray]:
    """计算图像色度直方图"""

    hsv = cv2.cvtColor(image.data(), cv2.COLOR_BGR2HSV)
    hue = hsv[..., 0]

    hist, edge = np.histogram(hue, bins, (0, 255), density=True)
    return hist, edge


def hist_vector(image: ImageNda, bins: int = 256) -> np.ndarray:
    """获取图像分布向量"""

    hist, edge = np.histogram(image.data(), bins, (0, 255), density=True)
    hist *= edge[1] - edge[0]
    assert hist.shape == (bins,)
    assert isclose(np.sum(hist), 1)
    return hist


'''
from matplotlib import pyplot as plt

def show_hist(hist: np.ndarray, edge: np.ndarray, width_factor: float = 0.7) -> None:
    """
绘制直方图
"""
    width = width_factor * (edge[1] - edge[0])
    center = (edge[:-1] + edge[1:]) / 2
    plt.bar(center, hist, align='center', width=width)
    plt.show()
"""


def calc_show_hist(image: ImageNda, bins: int = 128, width_factor: float = 0.7) -> None:
    """计算并绘制直方图"""

    hist, edge = np.histogram(image.data(), bins, (0, 255), density=True)

    show_hist(hist, edge, width_factor)
'''


def top_color_3c(_image: ImageNda, _ratio: float = 0.9) -> list[tuple[Color, float]]:
    """获取3通道图像前ratio颜色"""

    return []



from math import sqrt
from typing import Protocol, TypeVar, runtime_checkable

from jcx.m.number import Real

"""
泛型实验

Protocol问题:
- 不支持isinstance, 无法支持函数重载
- 提供的函数体无意义, 不能作为缺省实现
"""

T = TypeVar("T", int, float)


@runtime_checkable
class ISize2D(Protocol):
    """二维尺寸类型协议"""

    width: T
    height: T


def size_equal(s1: ISize2D, s2: ISize2D) -> bool:
    return s1.width == s2.width and s1.height == s2.height


def size_area(s: ISize2D) -> Real:
    """获取有二维尺寸对象面积"""
    return s.width * s.height


@runtime_checkable
class IHasSize2D(Protocol):
    """二维尺寸类型协议"""

    def size(self) -> ISize2D:
        pass


def area(s: ISize2D | IHasSize2D) -> Real:
    """获取有二维尺寸对象面积"""
    if isinstance(s, ISize2D):
        return size_area(s)
    elif isinstance(s, IHasSize2D):
        return size_area(s.size())
    raise "Invalid type"


def same_size(s1: IHasSize2D, s2: IHasSize2D) -> bool:
    """对象二维尺寸相等"""
    return size_equal(s1.size(), s2.size())


class IArea(Protocol):
    """面积类型协议"""

    def area(self) -> Real:
        pass


def dist_tuple(p1: tuple[float, float], p2: tuple[float, float]) -> float:
    """计算2D距离"""
    x1, y1 = p1
    x2, y2 = p2
    return sqrt((x1 - x2) ** 2 + (y1 - y2) ** 2)



#!/opt/ias/env/bin/python

import argparse
from pathlib import Path

from jvi.image.image_nda import ImageNda
from jvi.image.stat import value_map
from jvi.image.trace import trace_image

epilog = """
Examples:

    jvi-imp src.png dst.png
    
"""


def main():
    parser = argparse.ArgumentParser(description="图片处理工具", epilog=epilog)
    parser.add_argument("src", type=Path, help="来源图片")
    parser.add_argument("dst", type=Path, help="目的图片")
    parser.add_argument("-s", "--size", type=str, help="画布尺寸")
    parser.add_argument("-v", "--verbose", action="store_true", help="显示详细信息")
    opt = parser.parse_args()

    src_img = ImageNda.load(opt.src)

    print("source image shape:", src_img.shape())

    trace_image(src_img)
    m = value_map(src_img)
    print(f"{src_img} hist:")
    for v, p in m.items():
        print(f"\t{v}\t{int(p * 100)}%")


if __name__ == "__main__":
    main()



#!/opt/ias/env/bin/python

import argparse
from pathlib import Path

import sh
from jcx.sys import fs


def main() -> None:
    parser = argparse.ArgumentParser("批量转码工具")
    parser.add_argument("src_dir", type=Path, help="文件来源目录")
    parser.add_argument("dst_dir", type=Path, help="文件目标目录")
    parser.add_argument(
        "-e", "--src_ext", type=str, default=".mp4", help="来源视频文件扩展名"
    )
    parser.add_argument(
        "-d", "--dst_ext", type=str, default=".mp4", help="目标视频文件扩展名"
    )
    parser.add_argument(
        "-V", "--vcodec", type=str, default="h264", help="视频编码: h264/h265"
    )
    parser.add_argument("-a", "--acodec", type=str, default=None, help="音频编码codec")
    parser.add_argument(
        "-s", "--subtitle_ext", type=str, default=".srt", help="字幕文件扩展名"
    )
    parser.add_argument("--crop", type=str, default=None, help="裁切尺寸, 如: 960:720")
    parser.add_argument("-v", "--verbose", action="store_true", help="显示详细信息")
    opt = parser.parse_args()

    files = fs.files_in(opt.src_dir, opt.src_ext)

    print("src files:", len(files))

    vcodec_opts = {"h264": ["-c:v", "h264_nvenc"], "h265": ["-vcodec", "hevc"]}
    vcodec_opt = vcodec_opts[opt.vcodec]

    ff = sh.Command("/usr/local/bin/ffmpeg")
    for i, src in enumerate(files):
        print(f"#{i + 1} {src} ...")
        sub = src.with_suffix(opt.subtitle_ext)
        dst = opt.dst_dir / (src.stem + opt.dst_ext)
        if dst.is_file():
            print(f"  SKIP: {dst}")
            continue
        args = ["-i", src]
        if sub.is_file():
            args.extend(["-i", sub])
        if opt.crop:
            args.extend(["-vf", f"crop={opt.crop}"])
        args.extend(vcodec_opt)
        if opt.acodec:
            args.extend(["-acodec", opt.acodec])
        args.append(dst)
        ff(*args)


if __name__ == "__main__":
    main()



from typing import Optional

import cv2
import numpy as np
from jcx.m.number import Real1_2D
from jvi.drawing.color import Color
from jvi.geo.point2d import Points
from jvi.geo.rectangle import Rect
from jvi.geo.size2d import Size
from jvi.image.image_nda import ImageNda, ImageNdas
from jvi.image.util import make_mask, copy_mask


def resize(
    src: ImageNda,
    dst: Optional[ImageNda] = None,
    dst_size: Optional[Size] = None,
    scale: Optional[Real1_2D] = None,
) -> ImageNda:
    """缩放图片"""

    src_size = src.size()

    if dst_size is not None:
        assert isinstance(dst_size, Size)
        size = dst_size
    elif scale is not None:
        size = src_size.scale(scale).round()
    elif dst is not None:
        assert isinstance(dst, ImageNda)
        size = dst.size()
    else:
        raise RuntimeError("Invalid dst_size/scale argument")

    if dst:
        assert dst.same_channel_num(src)
    dst = cv2.resize(src.data(), size.round().to_tuple_int(), dst and dst.data())
    assert isinstance(dst, np.ndarray)
    return ImageNda(data=dst)


def resize_roi(
    src: ImageNda, rect: Rect, size: Optional[Size], dst: Optional[ImageNda]
) -> ImageNda:
    """获取图像区域并缩放到指定尺寸"""
    size = size or (dst and dst.size())
    roi: ImageNda = src.roi(rect)
    dst = resize(roi, dst, size)
    return dst


def resize_to_box(src: ImageNda, dst_box: ImageNda) -> ImageNda:
    """图像缩放, 保持宽高比, 放入目标图像内"""
    dst_size = src.size().scale_in(dst_box.size()).round()
    rect = Rect.from_cs(dst_box.size().center(), dst_size)
    resize(src, dst_box.roi(rect))
    return dst_box


def tile(src_images: ImageNdas, cols: int = 0, box_size: Size = Size()) -> ImageNda:
    """多个图像平铺合并成一个"""
    cols = cols or len(src_images)
    rows = (len(src_images) + cols - 1) // cols
    if not box_size:
        tile_size = Size()
        for im in src_images:
            tile_size |= im.size()
        box_size = tile_size.scale((cols, rows))

    dst_image = ImageNda(box_size)
    tiles = dst_image.to_tiles(cols, rows)

    for src, dst in zip(src_images, tiles):
        if src.channel_num() == 1:
            src = to_color(src)  # 灰度图转彩色
        resize_to_box(src, dst)

    return dst_image


def get_roi_image(
    src: ImageNda, region: Optional[Points], background: Optional[Color] = None
) -> ImageNda:
    """获取图片ROI图片"""
    size = src.size()
    region = region or Rect.one().vertexes()
    mask = make_mask(size, region)
    dst = ImageNda(size, color=background)
    copy_mask(src, mask, dst)
    return dst


def to_gray(src: ImageNda, dst: Optional[ImageNda] = None) -> ImageNda:
    """彩色图片抓换成灰度图"""
    assert src.channel_num() == 3

    dst_data = cv2.cvtColor(src.data(), cv2.COLOR_BGR2GRAY, dst and dst.data())
    return ImageNda(data=dst_data)


def to_color(src: ImageNda, dst: Optional[ImageNda] = None) -> ImageNda:
    """灰度图转为彩色图片"""
    assert src.channel_num() == 1

    dst_data = cv2.cvtColor(src.data(), cv2.COLOR_GRAY2BGR, dst and dst.data())
    return ImageNda(data=dst_data)


def merge(channels: ImageNdas) -> ImageNda:
    """合并通道成图像"""
    cs = [c.data() for c in channels]
    return ImageNda(data=cv2.merge(cs))


def split(image: ImageNda) -> ImageNdas:
    """合并通道成图像"""
    cs = cv2.split(image.data())
    return [ImageNda(data=c) for c in cs]


def laplacian_edge(src: ImageNda, dst: Optional[ImageNda] = None) -> ImageNda:
    """获取拉普拉斯边缘图"""
    lap = cv2.Laplacian(src.data(), cv2.CV_16S)  # CV_8U 抛弃负数, 不可用
    lap = np.absolute(lap).astype("uint8")
    im = ImageNda(data=lap)
    if dst is None:
        dst = im
    else:
        im.copy_to(dst)
    return dst


def blur(
    src: ImageNda, dst: Optional[ImageNda] = None, ksize: tuple[int, int] = (3, 3)
) -> ImageNda:
    """图像模糊"""
    if dst is not None:
        assert src.same_shape_type(dst)
        dst = cv2.blur(src.data(), ksize, dst.data())
    else:
        dst = cv2.blur(src.data(), ksize)
    assert isinstance(dst, np.ndarray)
    return ImageNda(data=dst)


def erode(
    src: ImageNda, dst: Optional[ImageNda] = None, ksize: tuple[int, int] = (3, 3)
) -> ImageNda:
    """图像腐蚀"""
    kernel = np.ones(ksize, np.uint8)
    if dst is not None:
        assert src.same_shape_type(dst)
        dst = cv2.erode(src.data(), kernel, dst.data())
    else:
        dst = cv2.erode(src.data(), kernel)
    assert isinstance(dst, np.ndarray)
    return ImageNda(data=dst)


def dilate(
    src: ImageNda, dst: Optional[ImageNda] = None, ksize: tuple[int, int] = (3, 3)
) -> ImageNda:
    """图像碰撞"""
    kernel = np.ones(ksize, np.uint8)
    if dst is not None:
        assert src.same_shape_type(dst)
        dst = cv2.dilate(src.data(), kernel, dst.data())
    else:
        dst = cv2.dilate(src.data(), kernel)
    assert isinstance(dst, np.ndarray)
    return ImageNda(data=dst)


def threshold(
    src: ImageNda,
    thresh: int,
    dst: Optional[ImageNda] = None,
    max_value: int = 255,
    thr_type: int = cv2.THRESH_BINARY,
) -> ImageNda:
    """图像二值化, x > thresh"""
    if dst is not None:
        assert src.same_shape_type(dst)
        r, dst = cv2.threshold(src.data(), thresh, max_value, thr_type, dst.data())
    else:
        r, dst = cv2.threshold(src.data(), thresh, max_value, thr_type)
    assert isinstance(dst, np.ndarray)
    return ImageNda(data=dst)


def threshold_range(
    src: ImageNda,
    thr_range: tuple[int, int],
    dst: Optional[ImageNda] = None,
    max_value: int = 255,
) -> ImageNda:
    """图像二值化, thr_range[0] < x <= thr_range[1]"""
    assert src.channel_num() == 1, "输入图像必须为灰度图"
    dst = threshold(
        src, thr_range[1], dst, 0, cv2.THRESH_TOZERO_INV
    )  # 高过阈值的部分设置为0
    dst = threshold(dst, thr_range[0], dst, max_value, cv2.THRESH_BINARY)
    return dst


def equalize_hist(src: ImageNda, dst: Optional[ImageNda] = None) -> ImageNda:
    """图像腐蚀"""
    data_dst = None
    if dst is not None:
        assert src.same_shape_type(dst)
        data_dst = dst.data()

    match src.channel_num():
        case 1:
            data_dst = cv2.equalizeHist(src.data(), data_dst)
        case 3:
            yuv = cv2.cvtColor(src.data(), cv2.COLOR_BGR2YUV)
            yuv[:, :, 0] = cv2.equalizeHist(yuv[:, :, 0])
            data_dst = cv2.cvtColor(yuv, cv2.COLOR_YUV2BGR)
        case _:
            raise RuntimeError("Invalid image channel number")
    assert isinstance(data_dst, np.ndarray)
    return ImageNda(data=data_dst)


def alpha_blend(im_src: ImageNda, im_dst: ImageNda, alpha: ImageNda) -> None:
    """图像alpha混合"""
    assert im_src.same_shape_type(im_dst)
    assert im_src.channel_at(0).same_shape(alpha)

    a = alpha.data() / 255
    b = 1 - a

    for i in range(im_src.channel_num()):
        c1 = im_src.channel_at(i).data()
        c2 = im_dst.channel_at(i).data()
        c2[:, :] = c1 * a + c2 * b


def canny(
    src: ImageNda, thr1: int = 100, thr2: int = 200, dst: Optional[ImageNda] = None
) -> ImageNda:
    if dst is not None:
        assert dst.channel_num() == 1
        assert dst.same_size(src)

    if src.channel_num() == 3:
        gray = to_gray(src)
    elif src.channel_num() == 1:
        gray = src
    else:
        raise Exception("not implemented")

    dst_data = cv2.Canny(gray.data(), thr1, thr2, dst and dst.data())
    return ImageNda(data=dst_data)



from copy import copy
from random import uniform
from typing import Any, Protocol, Self

from jcx.m.number import align_down, Real
from pydantic import BaseModel
from rustshed import Option, Null, Some

from jvi.geo import to_zero
from jvi.geo.point2d import Point, Points
from jvi.geo.size2d import Size


class Rect(BaseModel):
    """长方形"""

    x: float = 0
    y: float = 0
    width: float = 0
    height: float = 0

    @classmethod
    def new(cls, x: float, y: float, width: float, height: float) -> Self:
        """长方形-构造函数"""
        return Rect(x=x, y=y, width=width, height=height)

    @classmethod
    def zero(cls) -> Self:
        """长方形-空"""
        return Rect.new(0, 0, 0, 0)

    @classmethod
    def one(cls) -> Self:
        """长方形-归一化全图"""
        return Rect.new(0.0, 0.0, 1.0, 1.0)

    @classmethod
    def from_size(cls, s: Size) -> Self:
        """长方形-以Size构建"""
        return Rect.new(0, 0, s.width, s.height)

    @classmethod
    def from_ps(cls, p: Point, s: Size) -> Self:
        """长方形-以左上角和尺寸构建"""
        return Rect(x=p.x, y=p.y, width=s.width, height=s.height)

    @classmethod
    def from_ltrb(cls, p1: Point, p2: Point) -> Self:
        """长方形-以左上右下两点构建"""
        return Rect.new(p1.x, p1.y, p2.x - p1.x, p2.y - p1.y)

    @classmethod
    def from_cs(cls, center: Point, size: Size) -> Self:
        """长方形-以中心坐标+尺寸构建"""
        return Rect.from_cs_list(center.x, center.y, size.width, size.height)

    @classmethod
    def from_cs_list(cls, cx: float, cy: float, width: float, height: float) -> Self:
        """长方形-以中心坐标+尺寸构建"""
        return Rect.new(
            to_zero(cx - width / 2), to_zero(cy - height / 2), width, height
        )

    @classmethod
    def from_oi(cls, outer: Size, inner: Size) -> Self:
        """以外部长方形的中心为中心-以及内部尺寸为尺寸+尺寸构建"""
        center = Point(x=outer.width / 2, y=outer.height / 2)
        return Rect.from_cs(center, inner)

    def cs(self) -> tuple[Point, Size]:
        """获取中心+尺寸"""
        return self.center(), self.size()

    def cs_list(self) -> list[Real]:
        """获取中心+尺寸"""
        return [
            self.x + self.width / 2,
            self.y + self.height / 2,
            self.width,
            self.height,
        ]

    @classmethod
    def from_ltrb_list(cls, ll: list) -> Self:
        """长方形-以左上右下两点坐标列表构建，即：[x1, y1, x2, y2]"""
        return Rect.new(ll[0], ll[1], ll[2] - ll[0], ll[3] - ll[1])

    @classmethod
    def try_bounding(cls, points: Points) -> Option[Self]:
        """点集的外包长方形"""
        if len(points) < 1:
            return Null
        p1 = copy(points[0])
        p2 = copy(p1)
        for p in points:
            p1.x = min(p1.x, p.x)
            p1.y = min(p1.y, p.y)
            p2.x = max(p2.x, p.x)
            p2.y = max(p2.y, p.y)
        return Some(Rect.from_ltrb(p1, p2))

    @classmethod
    def bounding(cls, points: Points) -> Self:
        """点集的外包长方形"""
        return cls.try_bounding(points).unwrap()

    def __bool__(self) -> bool:
        """判定是否存在非零属性"""
        return self != Rect()

    def center(self) -> Point:
        """获取中心坐标"""
        return Point(x=self.x + self.width / 2, y=self.y + self.height / 2)

    def center_rect(self, size: Size) -> Self:
        """获取同心矩形"""
        return Rect.from_cs(self.center(), size)

    def size(self) -> Size:
        """获取尺寸"""
        return Size(width=self.width, height=self.height)

    def area(self) -> Real:
        """获取面积"""
        return self.width * self.height

    def aspect_ratio(self) -> Real:
        """获取宽高比"""
        return self.width / self.height

    def right(self) -> Real:
        """获取右侧值"""
        return self.x + self.width

    def bottom(self) -> Real:
        """获取底部值"""
        return self.y + self.height

    def lt(self) -> Point:
        """获取左上点"""
        return Point(x=self.x, y=self.y)

    def rt(self) -> Point:
        """获取右上点"""
        return Point(x=self.x + self.width, y=self.y)

    def rb(self) -> Point:
        """获取右下点"""
        return Point(x=self.x + self.width, y=self.y + self.height)

    def lb(self) -> Point:
        """获取左下点"""
        return Point(x=self.x, y=self.y + self.height)

    def ltrb(self) -> tuple[Point, Point]:
        """获取左上右下两点坐标"""
        return self.lt(), self.rb()

    def ltrb_tuple(self) -> tuple[tuple[float, float], tuple[float, float]]:
        """获取左上右下两点坐标 - tuple 表示"""
        return self.lt().to_tuple(), self.rb().to_tuple()

    def ltrb_tuple_int(self) -> tuple[tuple[int, int], tuple[int, int]]:
        """获取左上右下两点坐标 - tuple 表示"""
        return self.lt().to_tuple_int(), self.rb().to_tuple_int()

    def ltrb_list(self) -> list[Real]:
        """获取左上右下两点坐标列表"""
        return [self.x, self.y, self.x + self.width, self.y + self.height]

    def to_tuple_int(self) -> tuple[int, int, int, int]:
        """获取数据的tuple表示"""
        return int(self.x), int(self.y), int(self.width), int(self.height)

    def vertexes(self) -> Points:
        """获取四个定点坐标"""
        return [self.lt(), self.rt(), self.rb(), self.lb()]

    def clone(self) -> Self:
        """克隆对象"""
        return copy(self)

    def normalize_me(self, size: Size) -> None:
        """绝对坐标归一化"""
        self.x /= size.width
        self.y /= size.height
        self.width /= size.width
        self.height /= size.height

    def normalize(self, size: Size) -> Self:
        """获取绝对坐标归一化"""
        r = copy(self)
        r.normalize_me(size)
        return r

    def is_normalized(self) -> bool:
        """判断坐标是否被归一化"""
        p1, p2 = self.ltrb()
        return p1.is_normalized() and p2.is_normalized()

    def absolutize_me(self, size: Size) -> None:
        """归一化坐标绝对化"""
        self.x = round(self.x * size.width)
        self.y = round(self.y * size.height)
        self.width = round(self.width * size.width)
        self.height = round(self.height * size.height)

    def absolutize(self, size: Size) -> Self:
        """获取绝对坐标归一化对象"""
        r = copy(self)
        r.absolutize_me(size)
        return r

    def to_origin(self) -> Self:
        """产生新矩形: 左上角移动到原点"""
        return Rect.from_size(self.size())

    def scale_me(self, n: Real) -> None:
        """缩放指定的倍数"""
        self.x *= n
        self.y *= n
        self.width *= n
        self.height *= n

    def scale(self, n: Real) -> Self:
        """获取缩放指定的倍数的新对象"""
        r = self.clone()
        r.scale_me(n)
        return r

    def dilate_me(self, n: Real) -> None:
        """向四边膨胀指定值"""
        self.x -= n
        self.y -= n
        self.width += n * 2
        self.height += n * 2

    def dilate(self, n: Real) -> Self:
        """向四边膨胀指定值"""
        r = copy(self)
        r.dilate_me(n)
        return r

    def erode_me(self, n: Real) -> None:
        """向四边腐蚀指定值"""
        self.dilate_me(-n)

    def erode(self, n: Real) -> Self:
        """向四边腐蚀指定值"""
        return self.dilate(-n)

    def round_me(self) -> None:
        """近似成整数"""
        self.x = round(self.x)
        self.y = round(self.y)
        self.width = round(self.width)
        self.height = round(self.height)

    def round(self) -> Self:
        """近似成整数"""
        r = copy(self)
        r.round_me()
        return r

    def add_offset(self, offset: Point) -> None:
        """增加偏移量"""
        self.x += offset.x
        self.y += offset.y

    def contains_point(self, p: Point) -> bool:
        """判定矩形是否包含点"""
        rb = self.rb()
        return self.x <= p.x < rb.x and self.y <= p.y < rb.y

    def contains_rect(self, r: Self) -> bool:
        """判定矩形是否包含矩形"""
        p = r.rb()
        rb = self.rb()
        return (
            self.contains_point(r.lt())
            and self.x <= p.x <= rb.x
            and self.y <= p.y <= rb.y
        )

    def contains(self, v: Any) -> bool:
        """判定矩形是否包含对象"""
        if isinstance(v, Point):
            return self.contains_point(v)
        if isinstance(v, Rect):
            return self.contains_rect(v)
        return False

    def intersect(self, r: Self) -> Self:
        """获取矩形交集矩形"""
        a, b = self.ltrb()
        c, d = r.ltrb()
        x1 = max(a.x, c.x)
        y1 = max(a.y, c.y)
        x2 = min(b.x, d.x)
        y2 = min(b.y, d.y)

        return Rect.from_ltrb_list([x1, y1, x2, y2])

    def unite(self, r: Self) -> Self:
        """获取矩形并集矩形"""
        a, b = self.ltrb()
        c, d = r.ltrb()
        x1 = min(a.x, c.x)
        y1 = min(a.y, c.y)
        x2 = max(b.x, d.x)
        y2 = max(b.y, d.y)

        return Rect.from_ltrb_list([x1, y1, x2, y2])

    def iou(self, other: Self) -> float:
        """计算两个长方形交并比"""
        s1 = max(0, self.intersect(other).area())
        s2 = self.area() + other.area() - s1
        return s1 / s2

    def to_tiles(
        self,
        cols: int = 0,
        rows: int = 0,
        size: Size = Size(),
        need_round: bool = False,
    ) -> list[Self]:
        """把长方形切成瓦片"""
        assert (not cols) == (not rows)  # col/row 存在性一致
        assert (not cols) != (not size)  # col/size 存在性不同

        if cols:
            dx = self.width / cols
            dy = self.height / rows
        else:
            assert size
            dx = size.width
            dy = size.height
            cols = int(self.width / dx)
            rows = int(self.height / dy)

        rs = []
        for y in range(rows):
            for x in range(cols):
                r = Rect(x=self.x + x * dx, y=self.y + y * dy, width=dx, height=dy)
                if need_round:
                    r.round_me()
                rs.append(r)
        return rs

    def to_center_tiles(self, size: Size) -> list[Self]:
        """把长方形切成指定尺寸瓦片, 中心对称, 超出边界部分舍弃"""
        assert not self.is_normalized()
        assert not size.is_normalized()
        w = align_down(int(self.width), int(size.width))
        h = align_down(int(self.height), int(size.height))
        r = Rect.from_cs(self.center(), Size(width=w, height=h))
        return r.to_tiles(size=size)


type Rects = list[Rect]
"""长方形组"""


class PHasRect(Protocol):
    """可获取Rect类型"""

    def rect(self) -> Rect:
        """获取对象Rect"""
        pass


def letterbox(src: Size, dst: Size) -> Rect:
    """信封盒，把一个尺寸不改变宽高比，映射到新尺寸中的一个区域"""
    r = min(dst.width / src.width, dst.height / src.height)
    s = src.scale(r)
    return Rect.from_cs_list(dst.width / 2, dst.height / 2, s.width, s.height).round()


def center_tile_rect(size: Size, tile_side: int, row: int, col: int) -> Rect:
    """ "计算瓦片所在矩形区域"""

    x = (size.width - col * tile_side) / 2
    y = (size.height - row * tile_side) / 2

    return Rect(x=x, y=y, width=col * tile_side, height=row * tile_side)


def point_trans(p: Point, r: Rect) -> Point:
    """点坐标变换到目标区域内归一化坐标"""
    x = (p.x - r.x) / r.width
    y = (p.y - r.y) / r.height
    return Point(x=x, y=y)


def points_trans(ps: Points, r: Rect) -> Points:
    """点集坐标变换到目标区域内归一化坐标"""
    return [point_trans(p, r) for p in ps]


def random_point(range_: Size | Rect) -> Point:
    """获取指定范围内的随机点"""
    if isinstance(range_, Size):
        r = Rect.from_size(range_)
    elif isinstance(range_, Rect):
        r = range_
    else:
        raise "Invalid range type"
    x = uniform(r.x, r.right())
    y = uniform(r.y, r.bottom())
    return Point(x=x, y=y)



# from jvi.geo.point2d import Point
# from jvi.geo.rectangle import Rect
# from jvi.geo.size2d import Size


def is_normalized(v: float) -> bool:
    """判断坐标是否被归一化, 允许稍微越界, 这在绘图中常见"""
    return -0.01 <= v <= 1.01


def to_zero(v: float) -> float:
    """接近零的数据转为零"""
    return v if abs(v) > 0.00001 else 0



#!/opt/ias/env/bin/python

import cv2
import numpy as np


def main() -> None:
    file = "/home/jiang/ws/trash/sewage/s1/1/10-40-29.003.jpg"
    img = cv2.imread(file)
    gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)

    ret, binary = cv2.threshold(gray, 80, 255, cv2.THRESH_BINARY_INV)  # 阈值处理

    cv2.imshow("binary", binary)
    cv2.waitKey(0)
    cv2.destroyWindow("binary")
    contours, hierarchy = cv2.findContours(
        binary, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE
    )  # 查找轮廓
    print(len(contours))

    for i in range(len(contours)):
        area = cv2.contourArea(contours[i])
        if area < 1000:
            continue
        print("area:", area)
        cnt = contours[i]
        approx1 = cv2.approxPolyDP(cnt, 13, True)  # 拟合精确度
        print("approx1:", approx1.shape, type(approx1), approx1.dtype)
        approx2 = approx1.reshape(-1, 2)
        print("approx2:", approx2.shape, type(approx2), approx2.dtype)
        for p in approx2:
            print(p)
        img = cv2.polylines(img, [approx1], True, (255, 0, 255), 2)

    cv2.imshow("approxPolyDP1", img)

    cv2.waitKey(0)
    cv2.destroyAllWindows()


if __name__ == "__main__":
    main()



from dataclasses import dataclass
from pathlib import Path
from typing import Optional, TypeAlias

import cv2
import numpy as np
from jcx.sys.fs import files_in
from jvi.drawing.color import LIME
from jvi.geo.gp import dist_tuple
from jvi.geo.rectangle import Rect, Rects
from jvi.geo.size2d import Size, SIZE_HD
from jvi.image.image_nda import ImageNda
from jvi.image.proc import resize
from loguru import logger

DMatches: TypeAlias = list[cv2.DMatch]
"""匹配点集"""
DMatchMat: TypeAlias = list[DMatches]
"""匹配矩阵, 每个点匹配一个以上候选点, 但前只选择一个候补匹配点"""

KeyPoints: TypeAlias = tuple[cv2.KeyPoint, ...]
"""关键点集"""
DMatchPair: TypeAlias = tuple[cv2.DMatch, cv2.DMatch]
"""匹配点对"""

DMatchPairs: TypeAlias = tuple[DMatchPair, ...]
"""匹配点集合, 这里是257维"""


@dataclass(frozen=True)
class ImageMatchParams:
    """匹配参数"""

    nfeatures: int = 256
    """匹配的特征数上限"""
    min_match_count: int = 4
    """最少匹配特征数，低于该值表示完全失配"""
    max_top2_ratio: float = 0.7
    """最优/次优匹配点匹配距离比值，越小则最优匹配越可信，该特征越好"""
    size: Size = Size(224, 224)
    """图像尺寸"""


class ImageMatcher:
    """图像匹配器"""

    def __init__(self, params: ImageMatchParams = ImageMatchParams()):
        self.params = params
        self.detector = cv2.SIFT_create(nfeatures=params.nfeatures)
        # self.detector = cv2.ORB_create(nfeatures=params.nfeatures)

        flann_index_kdtree = 0
        index_params = dict(algorithm=flann_index_kdtree, trees=5)
        search_params = dict(checks=50)

        self.flann = cv2.FlannBasedMatcher(index_params, search_params)

        self.im1 = ImageNda(self.params.size)
        self.im2 = ImageNda(self.params.size)

    def match(
        self, image1: ImageNda, image2: ImageNda, match_im: Optional[ImageNda] = None
    ) -> float:
        """匹配两幅图像，获取平均匹配距离(相对)"""
        size = self.params.size
        resize(image1, self.im1)
        resize(image2, self.im2)

        kp1, des1 = self.detector.detectAndCompute(self.im1.data(), None)
        kp2, des2 = self.detector.detectAndCompute(self.im2.data(), None)

        try:
            matches = self.flann.knnMatch(des1, des2, k=2)  # 取两个点，方便比较
        except cv2.error:
            # 因为图像内容突变或平坦触发异常，暂时无解
            # 参考：http://5.9.10.113/66064786/opencv-python-sift-crashing
            logger.error("flann.knnMatch error")
            return 0

        good_matches = self.get_good_matches(matches)
        # print('good:', len(good))

        dist = self.match_dist(good_matches, kp1, kp2, size.width)

        if match_im is not None:
            self.draw_match_image(dist, good_matches, kp1, kp2, match_im, size)

        return dist  # TODO: 归一化，将就

    def get_good_matches(self, matches: DMatchPairs) -> DMatchMat:
        """过滤掉两个候选匹配差异不大的情况，减少错配"""
        good_matches = []
        for m, n in matches:
            if m.distance < self.params.max_top2_ratio * n.distance:
                good_matches.append([m])
        return good_matches

    def draw_match_image(
        self,
        dist: float,
        good: DMatches,
        kp1: KeyPoints,
        kp2: KeyPoints,
        match_im: ImageNda,
        size: Size,
    ) -> None:
        """画出匹配图, 用于调试"""
        out_size = match_im.size()
        in_size = size.scale((2, 1))
        r = Rect.from_oi(out_size, in_size)
        img_center = match_im.roi(r)
        cv2.drawMatchesKnn(
            self.im1.data(), kp1, self.im2.data(), kp2, good, img_center.data()
        )
        label = "matches=%d dist=%d" % (len(good), dist)
        cv2.putText(match_im.data(), label, (8, 16), 0, 0.5, LIME.bgr())

    def match_dist(
        self, good_matches: DMatchMat, kp1: KeyPoints, kp2: KeyPoints, max_dist: float
    ) -> float:
        """计算良好匹配的距离"""
        dist = max_dist
        if len(good_matches) > self.params.min_match_count:
            dists = []
            for m in good_matches:
                d = dist_tuple(kp1[m[0].queryIdx].pt, kp2[m[0].trainIdx].pt)
                dists.append(d)
            dists = sorted(dists)
            dist = dists[len(dists) // 2]  # 取中位数作为距离
        return dist

    def match_tiles(
        self, image1: ImageNda, image2: ImageNda, tiles: Rects
    ) -> list[float]:
        """按瓦片区域分块匹配"""

        dists = []
        for r in tiles:
            # print('r:', r)
            im1 = image1.roi(r)
            im2 = image2.roi(r)
            # rectangle(self.img_det, r, PINK)
            match_img = np.zeros_like(image2.data())  # FIXME
            d = self.match(im1, im2, match_img.data())
            dists.append(round(d, 2))
            # cv2.imshow('matches', match_img)
            # cv2.waitKey(0)
        return dists



#!/opt/ias/env/bin/python

import argparse
from pathlib import Path

from jcx.text.txt_json import load_json, save_json
from jvi.drawing.color import COLORS7
from jvi.drawing.shape import grid
from jvi.geo.size2d import size_parse
from jvi.image.tile import PanoramaInfo, divide_sparse
from jvi.image.io import trace_image, save_image


def main():
    parser = argparse.ArgumentParser(description="瓦片全景查看工具")
    parser.add_argument("panorama", type=Path, help="匹配全景文件")
    parser.add_argument(
        "-n", "--no-background", action="store_true", help="禁用背景图片"
    )
    parser.add_argument("-d", "--divide-size", type=str, help="切分尺寸，比如：4x4")
    parser.add_argument("-o", "--output", type=Path, help="图输出文件/目录")
    parser.add_argument(
        "-g", "--grid_thickness", type=int, default=None, help="绘制网格的线宽"
    )
    parser.add_argument("-v", "--verbose", action="store_true", help="显示详细信息")
    opt = parser.parse_args()

    pano_info = load_json(opt.panorama, PanoramaInfo)
    assert pano_info, "瓦片全景信息加载失败"

    img_pano = pano_info.load_tiles(opt.no_background)
    print("pano shape:", img_pano.shape)
    if opt.grid_thickness:
        grid(
            img_pano,
            pano_info.tile_size(),
            colors=COLORS7,
            thickness=opt.grid_thickness,
        )
    trace_image(img_pano)

    if opt.output:
        print("图片保存为：", opt.output)
        save_image(img_pano, opt.output)
        if opt.divide_size:
            size = size_parse(opt.divide_size)
            out_dir = Path(opt.output.stem)
            sparse_info = divide_sparse(
                img_pano, pano_info.tiles, out_dir, size.width, size.height
            )
            sparse_info.background = str(opt.output)
            out_file = out_dir.with_suffix(".json")
            print("稀疏全景信息信息保存到：", out_file)
            save_json(sparse_info, out_file)


if __name__ == "__main__":
    main()



#!/opt/ias/env/bin/python

import argparse
from pathlib import Path

import cv2
from jvi.geo.rectangle import Rect
from jvi.geo.size2d import size_parse
from jvi.image.image_nda import ImageNda

epilog = """
Examples:

    jvi-resize src.png dst.png -c 256x256
    
"""


def main():
    parser = argparse.ArgumentParser(description="图片缩放工具", epilog=epilog)
    parser.add_argument("src", type=Path, help="来源图片")
    parser.add_argument("dst", type=Path, help="目的图片")
    parser.add_argument(
        "-c", "--crop_size", type=str, help="目标图片裁切尺寸(中心裁切)"
    )
    parser.add_argument("-s", "--scale", type=float, default=1, help="图片缩放比例")
    parser.add_argument(
        "-i",
        "--interpolation",
        type=int,
        default=1,
        help="差之方法: 1-INTER_LINEAR, 2-INTER_CUBIC",
    )

    parser.add_argument("-v", "--verbose", action="store_true", help="显示详细信息")
    opt = parser.parse_args()

    assert opt.interpolation in (1, 2)

    src = ImageNda.load(str(opt.src), cv2.IMREAD_UNCHANGED)  # BGR
    assert src is not None
    dst = cv2.resize(
        src, (0, 0), None, fx=opt.scale, fy=opt.scale, interpolation=opt.interpolation
    )

    dst_size = dst.size()

    if opt.crop_size:
        crop_size = size_parse(opt.crop_size)

        if dst_size.contains(crop_size):
            r = Rect.from_cs(dst_size.center(), crop_size)
            crop = dst.roi(r)
        else:
            crop = ImageNda(crop_size, src.shape[2])
            center = crop_size.center()
            r = Rect.from_cs(center, dst_size).round()
            roi = crop.roi(r)
            cv2.copyTo(dst, None, roi)
        dst = crop
        if opt.verbose:
            cv2.imshow("crop", crop)

    if opt.verbose:
        cv2.imshow("src", src)
        cv2.imshow("dst", dst)
        cv2.waitKey(0)

    cv2.imwrite(str(opt.dst), dst)
    cv2.destroyAllWindows()


if __name__ == "__main__":
    main()



from abc import ABC
from typing import Any

import cv2
from jcx.sys.fs import StrPath
from jcx.ui.key import Key
from jvi.drawing.color import RED, Color
from jvi.drawing.shape import polylines
from jvi.geo.point2d import Point, Points
from jvi.geo.size2d import Size, SIZE_HD
from jvi.image.image_nda import ImageNda
from jvi.image.proc import resize


def mouse_callback(event: int, x: int, y: int, flags: int, me: Any) -> None:
    if me:
        me.on_mouse(event, Point.new(x, y), flags)


class ImageWin(ABC):

    def __init__(self, title: str, size: Size = SIZE_HD):
        self._title = title
        cv2.namedWindow(self._title)
        cv2.setMouseCallback(self._title, mouse_callback, self)
        self._size = size
        self._canvas: ImageNda = ImageNda(size)
        self._background: ImageNda = ImageNda(size)
        self._pos = Point()

    def __del__(self) -> None:
        # cv2.setMouseCallback(self._title, None, None)
        cv2.destroyWindow(self._title)

    def size(self) -> Size:
        """获取窗口尺寸（客户区）"""
        return self._size

    def resize(self, size: Size) -> None:
        """改变窗口尺寸（客户区）"""
        if self._size != size:
            self._size = size
            self._canvas = ImageNda(size)

    def background(self) -> ImageNda:
        """获取背景图"""
        return self._background

    def background_size(self) -> Size:
        """获取窗口尺寸（客户区）"""
        return self._background.size()

    def set_pos(self, p: Point) -> None:
        cv2.moveWindow(self._title, int(p.x), int(p.y))

    def set_background(self, image: ImageNda) -> None:
        """设置背景"""
        self._background = image.clone()

    def set_backcolor(self, color: Color) -> None:
        """设置背景色"""
        self._background.set_to(color)

    def canvas(self) -> ImageNda:
        return self._canvas

    def snapshot(self, file: StrPath) -> None:
        """抓图"""
        assert self
        self._canvas.save(file)
        print(f"抓图保存: {file}")

    def refresh(self) -> None:
        """刷新窗口"""
        resize(self._background, self._canvas)
        self.on_draw(self._canvas, self._pos)
        cv2.imshow(self._title, self._canvas.data())

    def run(self, interval: int = 200) -> int:
        """运行消息循环"""
        key = -1
        while True:
            self.refresh()
            if key > 0:
                r = self.on_key(key)
                if r:
                    cv2.destroyAllWindows()
                    return r
            else:
                self.on_idle()
            key = cv2.waitKey(interval)

    @staticmethod
    def wait_key(delay: int = 0) -> int:
        """等待按键，BUG:在接收鼠标消息后，小键盘按键消息无法接收"""
        return cv2.waitKey(delay)

    def on_mouse_move(self, _pos: Point, _flags: int) -> None:
        """设置鼠标移动事件响应"""
        pass

    def on_left_button_down(self, _pos: Point, _flags: int) -> None:
        """设置鼠标左键按下事件响应"""
        pass

    def on_right_button_down(self, _pos: Point, _flags: int) -> None:
        """设置鼠标右键按下事件响应"""
        pass

    def on_draw(self, _canvas: ImageNda, _pos: Point) -> None:
        """设置窗口重绘事件响应"""
        pass

    def on_key(self, key: int) -> int:
        """设置窗口重绘事件响应"""
        if key == Key.ESC:
            return 1
        return 0

    def on_mouse(self, event: int, pos: Point, flags: int) -> None:
        self._pos = pos
        if event == cv2.EVENT_MOUSEMOVE:
            self.on_mouse_move(pos, flags)
        elif event == cv2.EVENT_LBUTTONDOWN:
            self.on_left_button_down(pos, flags)
        elif event == cv2.EVENT_RBUTTONDOWN:
            self.on_right_button_down(pos, flags)

    def on_idle(self) -> None:
        """但窗口空闲时完成的工作"""
        pass


class PolylinesWin(ImageWin):

    def __init__(self, title: str, size: Size):
        super().__init__(title, size)
        self.points: Points = []
        self.counter = 0

    def on_key(self, key: int) -> int:
        if key == Key.ESC:
            return 1
        elif key == Key.BACKSPACE:
            if self.points:
                self.points.pop()
        elif key > 0:
            print("key:", key)
        return 0

    def on_draw(self, canvas: ImageNda, _pos: Point) -> None:
        color = RED
        if self.points:
            polylines(canvas, self.points, color, 2)

    def on_left_button_down(self, p: Point, flags: int) -> None:
        # print(flags)
        self.points.append(p)
        if flags & cv2.EVENT_FLAG_CTRLKEY:
            print("CTRL")
        if flags & cv2.EVENT_FLAG_ALTKEY:
            print("ALT")
        if flags & cv2.EVENT_FLAG_SHIFTKEY:
            print("SHIFT")
        # cv2.setWindowTitle()

    def on_idle(self) -> None:
        self.counter += 1
        if self.counter % 20 == 0:
            print("#%d idle" % self.counter)



#!/opt/ias/env/bin/python

import argparse
from pathlib import Path

import arrow
from jcx.sys.fs import files_in, stem_append
from jvi.common import JVI_ASSERTS
from jvi.drawing.color import *
from jvi.drawing.shape import polylines, cross, rectangle
from jvi.geo.point2d import array_normalize
from jvi.geo.rectangle import Rect
from jvi.geo.size2d import Size
from jvi.geo.trans import points_cs_trans_in_rect
from jvi.image.infrared import *
from jvi.image.proc import resize
from jvi.image.struct import OscrDetector, OscrParams
from jvi.image.trace import trace_image

IR_ROI: Final = Rect(x=396.0, y=140.0, width=1060, height=800)
"""红外ROI"""

epilog = """
Examples:

    jvi-infrared src.png
    
"""


def main() -> None:
    parser = argparse.ArgumentParser(description="红外图片处理工具", epilog=epilog)
    parser.add_argument("image_file", type=Path, help="来源图片")
    parser.add_argument("-e", "--exp", type=str, default=".jpg", help="图片文件扩展名")
    parser.add_argument("-v", "--verbose", action="store_true", help="显示详细信息")
    opt = parser.parse_args()

    assert opt.image_file.exists(), f"图片文件不存在: {opt.image_file}"

    image = ImageNda.load(opt.image_file, cv2.IMREAD_COLOR)

    params = OscrParams((0, 80))
    # params = OscrParams()
    detector = OscrDetector(params)

    objects = detector.detect(image)

    image1 = ImageNda.load("/home/jiang/1/a.jpg", cv2.IMREAD_COLOR)
    resize(image, image1.roi(IR_ROI))

    for ob in objects:
        polygon = points_cs_trans_in_rect(ob.polygon, Rect.one(), IR_ROI)
        polygon = array_normalize(polygon, image1.size())
        polylines(image1, polygon, random_color())
        rectangle(image1, IR_ROI, random_color())

    trace_image(image1, "test")
    detector.debug_images()


def match_image() -> None:
    a = "/home/jiang/1/a.jpg"
    b = "/home/jiang/1/b.jpg"

    im_a = ImageNda.load(a)
    im_b = ImageNda.load(b)

    size_b = Size(1060, 800)
    center_b = im_a.size().center()
    center_b.x += -34
    center_b.y += -0

    rx = 0.828125
    ry = 1.111111

    r = Rect.from_cs(center_b, size_b)
    print("roi:", r)
    im_roi = im_a.roi(r)
    resize(im_b, im_roi)

    cross(im_a, im_a.size().center(), 15, BLUE, 5)
    cross(im_roi, im_roi.size().center(), 15, PURPLE, 5)

    trace_image(im_a, "test")

    print("area ratio:", im_roi.size().area() / im_a.size().area())


def show_color() -> None:
    """尝试利用彩虹调色板分类热度图"""
    rainbow_file = JVI_ASSERTS / "infrared/rainbow.jpg"
    palette = load_palette_hs(rainbow_file)

    folder = "/home/jiang/ws/trash/sewage/demo/rainbow/"
    files = files_in(folder, ".jpg")
    for file in files:
        im_src = ImageNda.load(file)
        now = arrow.now()
        im_dst = extract_image_ir(palette, im_src)
        print("duration:", arrow.now() - now)

        # trace_image(im_dst, 'dst')
        im_dst.save(stem_append(file, "dst"))


"""
    src_file = JIV_STATIC / 'infrared/wh1.jpg'
    src = ImageNda.load(src_file)
    src_1c = to_gray(src)
    dst = map_color(src_1c, rainbow_color) # 彩虹调色板

    print(src, dst)

    trace_images([src, dst], 'Infrared', box_size=Size(1920, 540))
    close_all_windows()
"""

if __name__ == "__main__":
    # main()
    # match_image()
    show_color()



#!/opt/ias/env/bin/python

import argparse
from pathlib import Path

import cv2
from jcx.sys.fs import files_in, real_exe_path

"""
图片超分辨率工具

不适合整合为resize工具，因为分辨率受限。

"""


def resize_file(model, src: Path, dst: Path, scale: int, verbose: bool):
    img_src = ImageNda.load(str(src))
    if img_src is None:
        print("[ERROR] load image fail: " + str(src))
        return
    print("upsample source image ...")
    img_dst = model.upsample(img_src)
    print("imwrite image to:", dst)
    cv2.imwrite(str(dst), img_dst)

    if verbose:
        img_linear = cv2.resize(
            img_src, None, fx=scale, fy=scale, interpolation=cv2.INTER_LINEAR
        )
        img_bicubic = cv2.resize(
            img_src, None, fx=scale, fy=scale, interpolation=cv2.INTER_CUBIC
        )

        cv2.imshow("img_src", img_src)
        cv2.imshow("img_dst", img_dst)
        cv2.imshow("img_linear", img_linear)
        cv2.imshow("img_bicubic", img_bicubic)

        cv2.waitKey(0)


def main():
    parser = argparse.ArgumentParser(description="图片超分辨率工具")
    parser.add_argument("src", type=Path, help="来源图片文件/目录")
    parser.add_argument("dst", type=Path, help="目的图片文件/目录")
    parser.add_argument(
        "-s", "--scale", type=int, default=2, help="图片放大倍数: 2,3,4"
    )
    parser.add_argument("-e", "--ext", type=str, help="目录内待转换文件扩展名")
    parser.add_argument("-a", "--algorithm", type=str, default="edsr", help="放缩算法")
    parser.add_argument("-v", "--verbose", action="store_true", help="显示详细信息")
    opt = parser.parse_args()

    model_dir = real_exe_path().parent.parent / "model/superres"
    models = {2: "EDSR_x2.pb", 3: "EDSR_x3.pb", 4: "EDSR_x4.pb"}

    if opt.scale not in models:
        print("[ERROR] 无效放大倍数")
        return -2

    model = cv2.dnn_superres.DnnSuperResImpl_create()
    model_file = model_dir / models[opt.scale]
    print("load model:", model_file)
    model.readModel(str(model_file))

    model.setModel(opt.algorithm, opt.scale)

    if opt.ext:
        files = files_in(opt.src, opt.ext)
        for src in files:
            dst = opt.dst / src.name
            resize_file(model, src, dst, opt.scale, opt.verbose)
    else:
        resize_file(model, opt.src, opt.dst, opt.scale, opt.verbose)

    cv2.destroyAllWindows()


if __name__ == "__main__":
    main()



from typing import Final, Optional

import cv2
import numpy as np
from PIL import Image as ImagePIL
from jcx.sys import fs
from loguru import logger
from rustshed import Result, Err, Ok

from jvi.image.image import *

type_map: Final = {
    DType.U8: np.uint8,
    DType.U16: np.uint16,
    DType.U32: np.uint32,
    DType.U64: np.uint64,
    DType.I8: np.int8,
    DType.I16: np.int16,
    DType.I32: np.int32,
    DType.I64: np.int64,
    DType.F16: np.float16,
    DType.F32: np.float32,
    DType.F64: np.float64,
}

name_map: Final = {
    "uint8": DType.U8,
    "uint16": DType.U16,
    "uint32": DType.U32,
    "uint64": DType.U64,
    "int8": DType.I8,
    "int16": DType.I16,
    "int32": DType.I32,
    "int64": DType.I64,
    "float16": DType.F16,
    "float32": DType.F32,
    "float64": DType.F64,
}


class ImageNda(Image):
    """图片: 由numpy.ndarray实现"""

    @classmethod
    def new_as(
        cls,
        image: Image,
        size: Optional[Size] = None,
        channel: int = 0,
        dtype=None,
        color: Optional[ImageColor] = None,
    ) -> Self:
        """创建图片，缺省参数与指定图片相同"""
        s = size or image.size()
        c = channel or image.channel_num()
        t = dtype or image.dtype()
        return cls(shape=(s.height, s.width, c), dtype=t, color=color)

    @classmethod
    def try_load(cls, path: StrPath, flag: int = -1) -> Result[Self, str]:
        """尝试加载图片"""
        data = cv2.imread(str(path), flag)
        if data is None:
            return Err(f"Load image fail: {path}")
        return Ok(cls(data=data))

    @classmethod
    def load(cls, path: StrPath, flag: int = -1) -> Self:
        """加载图片, 失败抛出异常"""
        return cls.try_load(path, flag).unwrap()

    def save(self, path: StrPath) -> bool:
        """保存图片"""
        fs.make_parents(path)
        r = cv2.imwrite(str(path), self._data)
        assert isinstance(r, bool)
        return r

    def __init__(
        self,
        size: Size = Size.zero(),
        channel: int = 3,
        shape: Optional[tuple] = None,
        dtype: DType = DType.U8,
        data: Optional[np.ndarray] = None,
        color: Optional[ImageColor] = None,
    ):
        """构造图像"""
        if data is not None:
            assert isinstance(data, np.ndarray)
            self._data = data
        else:
            assert bool(size) ^ (shape is not None)
            if size:
                if channel == 1:
                    shape = size.to_shape()
                else:
                    shape = size.to_shape3d_i(channel)
            dtype1 = type_map.get(dtype)
            assert dtype1 is not None
            assert shape
            self._data = np.zeros(shape, dtype=dtype1)
        # assert len(self._data.shape) == 3 # 单通道图片为2?
        assert self.dtype() is not None
        if color is not None:
            self.set_to(color)

    def __bool__(self) -> bool:
        """判断图像是否有效"""
        return self._data is not None

    def size(self) -> Size:
        """获取图像尺寸"""
        return Size.new(self._data.shape[1], self._data.shape[0])

    def channel_num(self) -> int:
        """获取通道数"""
        return self._data.shape[2] if len(self.shape()) > 2 else 1

    def channel_at(self, idx: int) -> Self:
        """获取图像指定通道"""
        return ImageNda(data=self._data[..., idx])

    def channels(self) -> list[Self]:
        """获取图像所有通道"""
        return [self.channel_at(i) for i in range(self.channel_num())]

    def shape(self) -> Shape3D:
        """获取图像形状"""
        return self._data.shape

    def dtype(self) -> DType:
        """获取数据类型"""
        return name_map.get(self._data.dtype.name)

    def clone(self) -> Self:
        """克隆图片对象"""
        return ImageNda(data=self._data.copy())

    def copy_to(self, other: Self) -> None:
        """复制数据到目标图像"""
        cv2.copyTo(self._data, None, other.data())

    def convert_to(self, other: Self) -> None:
        raise NotImplementedError()

    def set_to(self, color: ImageColor) -> None:
        """设置颜色"""
        if isinstance(color, Color):
            color = color.bgr()

        if isinstance(color, int | float):
            assert self.channel_num() == 1
        else:
            assert len(color) == self.channel_num()
        self._data[:] = color

    def roi(self, rect: Rect) -> Self:
        """获取感兴趣区域"""
        rect = rect.clone()
        if rect.is_normalized():
            rect = rect.absolutize(self.size())
        rect.round_me()
        a, b = rect.ltrb()
        return ImageNda(data=self._data[int(a.y) : int(b.y), int(a.x) : int(b.x)])

    def data(self) -> np.ndarray:
        """获取数据"""
        return self._data

    def __eq__(self, other: Self) -> bool:
        """判断是否相等"""
        return (self.data() == other.data()).all()

    def to_tiles(self, cols: int, rows: int) -> list[Self]:
        """图像切成块, 各个块与原图像共享数据"""
        rects = Rect.from_size(self.size()).to_tiles(cols, rows)
        return [self.roi(r) for r in rects]


ImageNdas: TypeAlias = list[ImageNda]
"""图像数组"""


def new_gray(size: Size, color: Real = 0, dtype: DType = DType.U8) -> ImageNda:
    """产生一个灰度图"""
    return ImageNda(size, 1, dtype=dtype, color=color)


def is_image(f: StrPath) -> bool:
    """判断文件是否是图片"""
    r = ImageNda.try_load(f)
    match r:
        case Err(_s):
            return False
    return True


def correct_image(f: StrPath) -> bool:
    """更正图片文件错误(CRC), FIXME: 好像没用"""
    r = ImageNda.try_load(f)
    match r:
        case Err(_s):
            image = ImagePIL.open(f)
            image.save(f)
            logger.info(f"Correct image: {f}")
            return False
    return True



from dataclasses import dataclass
from pathlib import Path
from typing import TypeAlias, Protocol, Sequence, TypeVar, Generic, List, Optional

from jcx.sys.fs import StrPath
from jvi.geo.point2d import Point
from jvi.geo.size2d import Size, SIZE_HD
from jvi.gui.image_viewer import ImageViewer
from jvi.image.image_nda import ImageNda
from rustshed import Option, Null


@dataclass
class PImageEntry(Protocol):
    """图片信息"""

    def get_image(self) -> ImageNda:
        """获取图片"""
        pass

    def image_file(self) -> Optional[Path]:
        """获取图片路径"""
        pass

    def draw_on(self, canvas: ImageNda, pos: Point) -> None:
        """把图片信息绘制在画板上"""
        pass


ImageEntries: TypeAlias = Sequence[PImageEntry]
"""图片信息集合"""

IE = TypeVar("IE", bound="PImageEntry")


class RecordViewer(ImageViewer, Generic[IE]):
    """记录图片查看窗口"""

    def __init__(self, title: str, size: Size = SIZE_HD):
        super().__init__(title, size)
        self._records: List[IE] = []

    def on_draw(self, canvas: ImageNda, pos: Point) -> None:
        """设置窗口重绘事件响应"""
        self.record().draw_on(canvas, pos)

    def image_at(self, index: int) -> ImageNda:
        """获取指定索引的图片"""
        return self._records[index].get_image()

    def cur_image_file(self) -> Optional[Path]:
        return self._records[self._index].image_file()

    def image_count(self) -> int:
        """获取图片总数"""
        return len(self._records)

    def set_records(self, records: List[IE]) -> None:
        """设置记录"""
        # self._records = sorted(records)
        self._records = records
        if self._records:
            self.change_background()

    def record(self) -> IE:
        """获取当前记录"""
        return self._records[self._index]


@dataclass
class FileRecord(PImageEntry):
    """文件记录"""

    path: Path
    """图片路径"""
    image: Option[ImageNda] = Null
    """图片对象, 内存中的图片数据"""

    def get_image(self) -> ImageNda:
        """加载图片"""
        return ImageNda.load(self.path)

    def image_file(self) -> Path:
        """获取图片路径"""
        return self.path

    def draw_on(self, canvas: ImageNda, _pos: Point) -> None:
        """把记录绘制在画板上"""
        # label = self.path.name
        # draw_boxf(canvas, Rect(0.25, 0.25, 0.5, 0.5), GREEN, label, 3)
        pass


FileRecords: TypeAlias = list[FileRecord]
"""文件记录列表"""


def load_dir_records(src_dir: StrPath) -> FileRecords:
    """加载目录下的图片信息记录"""
    src_dir = Path(src_dir)
    # print(src_dir)
    files = sorted(src_dir.rglob("*.jpg"))

    rs = [FileRecord(f) for f in files]
    return rs



#!/opt/ias/env/bin/python

import argparse
from pathlib import Path
from typing import Optional

from jcx.text.txt_json import save_json
from jvi.image.tile import panorama_from_dir


def main():
    parser = argparse.ArgumentParser(description="目录图片瓦片平铺工具")
    parser.add_argument("src_dir", type=Path, help="来源图片目录")
    parser.add_argument(
        "-p", "--panorama", type=Optional[Path], default=None, help="匹配全景文件"
    )
    parser.add_argument(
        "-e", "--ext", type=str, default=".png", help="来源图片文件扩展名"
    )
    parser.add_argument("-c", "--cols", required=True, type=int, help="瓦片列数")
    parser.add_argument("-r", "--rows", type=int, help="瓦片行数")
    parser.add_argument("-v", "--verbose", action="store_true", help="显示详细信息")
    opt = parser.parse_args()

    panorama = opt.panorama or opt.src_dir.with_suffix(".json")
    print("匹配信息输出到：", panorama)

    info = panorama_from_dir(
        opt.src_dir, cols=opt.cols, rows=opt.rows, ext=opt.ext, verbose=opt.verbose
    )
    save_json(info, panorama)


if __name__ == "__main__":
    main()



from typing import Protocol


class PPoint2D(Protocol):
    x: float
    y: float


class PSize2D(Protocol):
    width: float
    height: float


class PRect(Protocol):
    x: float
    y: float
    width: float
    height: float



#!/opt/ias/env/bin/python

import argparse
from pathlib import Path

from jcx.sys.fs import files_in
from jvi.drawing.color import YOLO_GRAY
from jvi.geo.rectangle import Rect
from jvi.geo.size2d import size_parse
from jvi.image.image_nda import ImageNda
from jvi.image.trace import trace_image

epilog = """
Examples:

    jvi-letterbox src_dir dst_dir -S 256x256
    
"""


def main() -> None:
    parser = argparse.ArgumentParser(description="图片嵌入工具", epilog=epilog)
    parser.add_argument("src_dir", type=Path, help="来源图片目录")
    parser.add_argument("dst_dir", type=Path, help="目的图片目录")
    parser.add_argument(
        "-S", "--dst_size", type=str, default="HD", help="目标图片裁切尺寸(中心裁切)"
    )
    parser.add_argument("-v", "--verbose", action="store_true", help="显示详细信息")
    opt = parser.parse_args()

    dst_size = size_parse(opt.dst_size)
    assert dst_size

    files = files_in(opt.src_dir, ".jpg")
    assert len(files)
    for src_file in files:
        src = ImageNda.load(src_file)
        assert src is not None
        dst = ImageNda(dst_size, color=YOLO_GRAY)
        center = dst_size.center()
        r = Rect.from_cs(center, src.size()).round()
        src.copy_to(dst.roi(r))
        dst_file = Path(opt.dst_dir, src_file.name)
        trace_image(dst, dst_file.name)
        dst.save(dst_file)


if __name__ == "__main__":
    main()



import sys
from copy import copy
from dataclasses import field, dataclass
from typing import Optional, Self

from pydantic import BaseModel

from jvi.geo.line_segment import LineSegments, LineSegment
from jvi.geo.point2d import Point, Points
from jvi.geo.rectangle import Rect
from jvi.geo.size2d import Size


class Polygon(BaseModel):
    """多边形"""

    vertexes: Points = field(default_factory=list)

    @classmethod
    def new(cls, vertexes: Points) -> Self:
        """创建多边形"""
        return Polygon(vertexes=vertexes)

    @classmethod
    def by(cls, shape: Rect):
        pass

    def __len__(self) -> int:
        """获取节点数"""
        return len(self.vertexes)

    def at(self, i: int) -> Point:
        """获取指定索引的顶点，可回绕"""
        i = (i + len(self)) % len(self)
        return self.vertexes[i]

    def borders(self) -> LineSegments:
        """获取边界线段集"""
        ss = []
        for i in range(len(self)):
            ss.append(LineSegment(self.at(i), self.at(i + 1)))
        return ss

    def insert_best(self, p: Point) -> int:
        """插入一个顶点，保证多边形离原有几点最近，且不自相交，返回插入位置"""

        if len(self) < 3:
            self.vertexes.append(p)
            return len(self) - 1

        best_i = -1
        best_d = sys.float_info.max
        for i in range(len(self)):
            if self._check(p, i) or self._check(p, i + 1):
                continue
            d = p.dist(self.at(i)) + p.dist(self.at(i + 1))
            if d < best_d:
                best_d = d
                best_i = i + 1
        assert best_i >= 0
        self.vertexes.insert(best_i, p)
        return best_i

    def del_neighbors(self, p: Point) -> None:
        """删除一个顶点的两个邻居"""
        i = self.vertexes.index(p)
        neighbors = [self.vertexes[i - 1], self.vertexes[(i + 1) % len(self.vertexes)]]
        for v in neighbors:
            if len(self.vertexes) > 1:
                self.vertexes.remove(v)

    def _check(self, p: Point, i: int) -> bool:
        """检查一点与指定顶点连线，是否与其他边相交"""
        p1 = self.at(i)
        line = LineSegment(p, p1)
        poly = self.polyline(i + 1, len(self) - 1)
        return line.intersects_polyline(poly)

    def polyline(self, start: int, size: int) -> Points:
        """从多边形上截取一段折线"""
        polygon = []
        for i in range(size):
            polygon.append(self.at(i + start))
        return polygon

    def is_normalized(self) -> bool:
        """判断坐标是否被归一化"""
        for p in self.vertexes:
            if not p.is_normalized():
                return False
        return True

    def normalize_me(self, size: Size) -> None:
        """绝对坐标归一化"""
        for p in self.vertexes:
            p.normalize_me(size)

    def normalize(self, size: Size) -> Self:
        """获取绝对坐标归一化"""
        p = copy(self)
        p.normalize_me(size)
        return p

    def absolutize_me(self, size: Size) -> None:
        """归一化坐标绝对化"""
        for p in self.vertexes:
            p.absolutize_me(size)
        self.round_me()

    def absolutize(self, size: Size) -> Self:
        """获取归一化坐标绝对化"""
        p = copy(self)
        p.absolutize_me(size)
        return p

    def round_me(self) -> None:
        """近似成整数"""
        for p in self.vertexes:
            p.round_me()

    def round(self) -> Self:
        """近似成整数"""
        s = copy(self)
        s.round_me()
        return s


def rect_polygon(rect: Rect) -> Optional[Polygon]:
    """长方形构造多边形"""
    return Polygon.new(rect.vertexes())


type Polygons = list[Polygon]
"""多边形集合"""



#!/opt/ias/env/bin/python

import argparse
from pathlib import Path
from typing import Optional

from jcx.sys.fs import files_in
from jcx.text.txt_json import save_json
from jvi.geo.rectangle import Rect
from jvi.image.tile import PanoramaInfo, TileInfo
from jvi.image.util import ndarray_size
from jvi.match.match_template import find_template_tile

"""
从原图中查找模板（覆盖匹配），图片位置输出到到全景图数据中(JSON格式)。
"""


def main():
    parser = argparse.ArgumentParser(description="瓦片图片匹配工具")
    parser.add_argument("src", type=Path, help="待匹配图片文件")
    parser.add_argument("template", type=Path, help="模板图片文件/目录")
    parser.add_argument(
        "-p", "--panorama", type=Optional[Path], default=None, help="匹配全景文件"
    )
    parser.add_argument("-e", "--ext", type=str, default=".png", help="图片文件扩展名")
    parser.add_argument("-a", "--align", type=int, default=8, help="坐标对齐")
    parser.add_argument("-v", "--verbose", action="store_true", help="显示详细信息")
    opt = parser.parse_args()

    flag = None  # cv2.IMREAD_UNCHANGED
    img_src = ImageNda.load(str(opt.src), flag)
    assert img_src is not None

    panorama = opt.panorama or opt.src.with_suffix(".json")

    print("匹配信息输出到：", opt.panorama)

    files = files_in(opt.template, opt.ext)
    pano_info = PanoramaInfo(ndarray_size(img_src), str(opt.src))

    for i, f in enumerate(files):
        print("开始搜索:", f.name)
        img_templ = ImageNda.load(str(f), flag)
        assert img_templ is not None
        size = ndarray_size(img_templ)

        pos, v = find_template_tile(img_src, img_templ)
        pos1 = pos.align_round(opt.align)
        msg = "" if pos == pos1 else "\talign"
        print(i, pos, v, msg)

        rect = Rect.from_ps(pos1, size)
        pano_info.tiles.append(TileInfo(str(f), rect))

        save_json(pano_info, panorama)


if __name__ == "__main__":
    main()



#!/opt/ias/env/bin/python

import argparse
from pathlib import Path

import cv2
from jvi.geo.size2d import size_parse


def main():
    parser = argparse.ArgumentParser(description="图片转视频工具")
    parser.add_argument(
        "src_dir", metavar="SRC_DIR", type=Path, help="来源图片所在目录"
    )
    parser.add_argument("dst_file", metavar="DST_FILE", type=Path, help="目的视频文件")
    parser.add_argument(
        "-p", "--pattern", type=str, default="*.jpg", help="图片文件模式(*.jpg)"
    )
    parser.add_argument("-s", "--size", type=str, default="HD", help="输出视频尺寸(HD)")
    parser.add_argument("-f", "--fps", type=float, default=5, help="视频帧率(5)")
    parser.add_argument(
        "-r", "--repeat", type=int, default=1, help="视频帧的重复次数(1)"
    )
    parser.add_argument("-v", "--verbose", action="store_true", help="显示详细信息")
    opt = parser.parse_args()

    size = size_parse(opt.size)

    fourcc_tab = {
        ".mp4": "mp4v",
        ".mjpg": "mjpg",
        ".webp": "webp",
    }  # require libwebp-dev
    # TODO: avc1/MJPG报错，https://stackoverflow.com/questions/52932157/ \
    #  opencv-ffmpeg-tag-0x34363268-h264-is-not-supported-with-codec/56723380

    fourcc = fourcc_tab.get(opt.dst_file.suffix)
    if not fourcc:
        print("[ERROR] 不支持的文件格式:", opt.dst_file.suffix)
        return -1
    fourcc = cv2.VideoWriter_fourcc(*fourcc)
    writer = cv2.VideoWriter(str(opt.dst_file), fourcc, opt.fps, size.to_tuple_int())

    files = opt.src_dir.rglob(opt.pattern)
    files = sorted(files)
    print("源图片目录：%s" % opt.src_dir)

    dst_img = None
    ok = 0
    err = 0
    for i, f in enumerate(files):
        src_img = ImageNda.load(str(f))  # BGR
        name = f.name
        if src_img is None:
            err += 1
            print(i, name, "\tERROR")
            continue
        else:
            ok += 1
            if opt.verbose:
                print(i, name, "\tOK")

        dst_img = cv2.resize(src_img, size.to_tuple_int(), dst_img)
        for _i in range(opt.repeat):
            writer.write(dst_img)

    print("\n压缩帧数：(%d/%d)\n" % (ok, ok + err))


if __name__ == "__main__":
    main()



from dataclasses import dataclass, field
from typing import Optional, TypeAlias

import cv2
import numpy as np
from jvi.drawing.color import random_color
from jvi.drawing.shape import polylines
from jvi.geo.point2d import Point, Points, array_normalize
from jvi.geo.size2d import Size
from jvi.image.image_nda import ImageNda
from jvi.image.proc import to_gray, threshold_range, erode, resize, dilate, threshold
from jvi.image.trace import trace_images
from jvi.image.util import mat_to_points


@dataclass
class BoxObject:
    """盒子目标, TODO"""

    id: int = 0
    """"目标ID"""
    polygon: Points = field(default_factory=list)
    """目标范围"""
    area: float = 0
    """目标面积"""

    def normalize_me(self, size: Size) -> None:
        """归一化自身坐标 & 面积"""
        self.polygon = array_normalize(self.polygon, size)
        self.area = self.area / size.area()


BoxObjects: TypeAlias = list[BoxObject]
"""盒子目标集定义"""


def find_polygons(
    mask: ImageNda,
    min_area: float = 0,
    min_area_ratio: float = 0,
    epsilon: int = 13,
    make_box: bool = False,
) -> BoxObjects:
    """
    图像中获取目标的外包多边形.

    Parameters
        -------
        mask: 二值化图像.

        min_area: 为最小面积.

        min_area_ratio: 为最小面积比率, 二者取其大

        epsilon: 多边形拟合步长

        make_box: True-拟合长方形, False-拟合多边形
    Returns
        -------
        out: 盒子目标集, 未归一化.

    """
    assert mask.channel_num() == 1
    full_area = mask.size().area()
    min_area = max(min_area, full_area * min_area_ratio)
    assert 0 <= min_area_ratio <= 1
    assert min_area >= 1
    # print('min_area:', min_area)

    contours, hierarchy = cv2.findContours(
        mask.data(), cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE
    )
    objects: BoxObjects = []
    for contour in contours:
        area = cv2.contourArea(contour)
        if area < min_area:
            continue
        if make_box:  # 拟合矩形
            rect = cv2.minAreaRect(contour)  # 输出浮点坐标
            box = cv2.boxPoints(rect)
            assert isinstance(box, np.ndarray)
            box1 = np.int0(box)
            assert isinstance(box1, np.ndarray)
            polygon = [Point(x, y) for x, y in box1]
        else:  # 拟合多边形
            approx1 = cv2.approxPolyDP(contour, epsilon, True)  # 拟合精确度
            polygon = mat_to_points(approx1)
        objects.append(BoxObject(0, polygon, area))
    return objects


@dataclass
class OscrParams:
    """多边形"""

    color_range: tuple[int, int] = (0, 0)
    """颜色范围, 默认使用OTSU自动阈值(目前全图范围效果不好, 可以实施ROI)"""
    ksize: tuple[int, int] = (7, 7)
    """开运算卷积核尺寸"""
    min_area_ratio: float = 1 / 64
    """目标最小面积比率, 计算得到的面积值会远小于感官, 该值不宜过大"""
    scale: float = 0.5
    """图像缩放比例"""


class OscrDetector:
    """特定颜色范围目标检测器(Objects with a Specific Color Range)"""

    def __init__(self, params: OscrParams = OscrParams()):
        self._color_range = params.color_range
        self._ksize = params.ksize
        self._min_area_ratio = params.min_area_ratio
        self._scale = params.scale

        self._im_color: Optional[ImageNda] = None  # 彩色图
        self._im_gray: Optional[ImageNda] = None  # 灰度图
        self._im_bin: Optional[ImageNda] = None  # 二值图
        self._objects: BoxObjects = []  # 检测到的目标

    def size(self) -> Size:
        """获取内部图像尺寸"""
        return self._im_color.size() if self._im_color else Size()

    def detect(self, image: ImageNda) -> BoxObjects:
        """灰度范围目标检测器, 返回归一化目标"""
        self._im_color = resize(image, self._im_color, scale=0.5)
        self._im_gray = to_gray(self._im_color, self._im_gray)

        if self._color_range == (0, 0):
            self._im_bin = threshold(
                self._im_gray,
                0,
                self._im_bin,
                255,
                cv2.THRESH_BINARY_INV + cv2.THRESH_OTSU,
            )
        else:
            self._im_bin = threshold_range(
                self._im_gray, self._color_range, self._im_bin
            )
        erode(self._im_bin, self._im_bin, self._ksize)
        dilate(self._im_bin, self._im_bin, self._ksize)

        objects = find_polygons(self._im_bin, min_area_ratio=self._min_area_ratio)
        for ob in objects:
            ob.normalize_me(self._im_bin.size())
        self._objects = objects
        return self._objects

    def debug_images(self) -> None:
        """显示内部图片, 用于调试"""
        assert self._im_color and self._im_gray and self._im_bin
        for ob in self._objects:
            polylines(self._im_color, ob.polygon, random_color())
        trace_images([self._im_color, self._im_gray, self._im_bin], "Infrared")



import cv2
import numpy as np
from PIL.Image import Image as PilImage, fromarray
from jvi.image.image_nda import ImageNda


def bgr_to_pil(bgr: np.ndarray) -> PilImage:
    """BGR图像转PIL图像"""
    rgb = cv2.cvtColor(bgr, cv2.COLOR_BGR2RGB)
    return fromarray(rgb)


def pil_to_bgr(pil: PilImage) -> np.ndarray:
    """PIL图像转BGR图像"""
    rgb = np.asarray(pil)
    bgr = cv2.cvtColor(rgb, cv2.COLOR_RGB2BGR)
    assert isinstance(bgr, np.ndarray)
    return bgr


def pil_to_nda(pil: PilImage) -> ImageNda:
    """PIL图像转ImageNda图像"""
    im = pil_to_bgr(pil)
    return ImageNda(data=im)



from dataclasses import dataclass, field
from pathlib import Path
from typing import Optional, TypeAlias

import cv2  # type: ignore
from jcx.sys.fs import files_in
from jvi.geo.point2d import Point
from jvi.geo.rectangle import Rect
from jvi.geo.size2d import Size
from jvi.image.image_nda import ImageNda


@dataclass
class TileInfo:
    """瓦片信息"""

    file: str
    """瓦片所在文件"""
    rect: Rect
    """瓦片所在位置"""


TileInfos: TypeAlias = list[TileInfo]


@dataclass
class PanoramaInfo:
    """全景图信息"""

    size: Size
    """全图尺寸"""
    background: Optional[str] = None
    """背景图"""
    tiles: list[TileInfo] = field(default_factory=list)
    """瓦片图列表"""

    def tile_size(self) -> Size:
        """获取瓦片尺寸"""
        return self.tiles[0].rect.size()

    def load_tiles(self, no_background: bool) -> ImageNda:
        """ "加载图片集合"""

        assert self.background
        pano = (
            ImageNda.load(self.background, cv2.IMREAD_UNCHANGED)
            if not no_background
            else None
        )
        tile_size = None
        print("开始加载图片:")
        for i, tile in enumerate(self.tiles):
            print("  #%d %s" % (i, tile.file))
            img = ImageNda.load(tile.file, cv2.IMREAD_UNCHANGED)  # FIXME:
            assert img is not None

            if pano is None:
                pano = ImageNda.new_as(img, size=self.size)  # 目标图是黑空图
            assert pano is not None
            assert self.size == pano.size()

            size = img.size()
            if tile_size is None:
                tile_size = size
            assert size == tile_size
            assert size == tile.rect.size()

            roi = pano.roi(tile.rect)
            cv2.copyTo(img, None, roi)

        return pano


def divide_sparse(
    image: ImageNda,
    tile_infos: TileInfos,
    dst_dir: Path,
    rows: int,
    cols: int,
    file_name_fmt: str = "%02d-%02d.png",
) -> PanoramaInfo:
    """把拼接好的图像分割稀疏块"""

    pano_size = image.size()
    pano_info = PanoramaInfo(pano_size, None)

    tile_rects = Rect.from_size(pano_size).to_tiles(rows=rows, cols=cols)
    print("开始切分稀疏瓦片图：")
    for i, rect in enumerate(tile_rects):
        ok = False
        for t in tile_infos:
            r = rect.intersect(t.rect)
            if r.size().positive():
                ok = True
                break
        if ok:
            x = i % cols
            y = i // rows
            file = dst_dir / (file_name_fmt % (x, y))
            print("  #%d %s" % (i, file))
            roi = image.roi(rect)
            roi.save(file)
            pano_info.tiles.append(TileInfo(str(file), rect))

    return pano_info


def panorama_from_dir(
    src_dir: Path, cols: int, rows: int = 0, ext: str = ".png", verbose=False
) -> PanoramaInfo:
    """从目录文件建立全景信息"""
    assert cols > 0

    files = files_in(src_dir, ext)
    assert len(files) > 0, "[ERROR] 目录中没有图片"

    rows = rows or (len(files) + cols - 1) // cols
    assert rows > 0

    pano_info = None
    tile_size = None

    for i, f in enumerate(files):
        img = ImageNda.load(str(f), cv2.IMREAD_UNCHANGED)
        assert img is not None, "[ERROR] 加载文件失败:"

        size = img.size()
        if not pano_info:
            tile_size = size
            pano_size = tile_size * Size(cols, rows)
            pano_info = PanoramaInfo(pano_size, None)
        assert size == tile_size

        x = i % cols
        y = i // rows
        pos = Point(x * size.width, y * size.height)
        rect = Rect.from_ps(pos, size)
        pano_info.tiles.append(TileInfo(str(f), rect))
        if verbose:
            print(i, pos)
    return pano_info



from pathlib import Path
from typing import Optional

import cv2
import imageio.v2 as iio
from jcx.sys.fs import StrPath, make_parents, files_in
from jvi.geo.size2d import Size
from jvi.image.image_nda import ImageNda, ImageNdas
from jvi.image.proc import resize
from jvi.image.trans import bgr_to_pil


# 参考: [PIL Image file formats](https://pillow.readthedocs.io/en/stable/handbook/image-file-formats.html)


def _save_images_cv(
    images: ImageNdas, path: StrPath, fps: float, size: Size = Size()
) -> None:
    """保存图片集合成动画/视频文件, 只有 'mp4v'勉强能用"""
    assert images
    path = Path(path)
    size = size or images[0].size()
    make_parents(path)

    fourcc_tab = {
        ".mp4": "avc1",
        ".mjpg": "mjpg",
        ".webp": "webp",
    }  # require libwebp-dev
    # TODO: avc1/MJPG报错，https://stackoverflow.com/questions/52932157/ \
    #  opencv-ffmpeg-tag-0x34363268-h264-is-not-supported-with-codec/56723380
    fourcc = fourcc_tab.get(path.suffix)
    assert fourcc, "不支持的文件格式"

    fourcc = cv2.VideoWriter_fourcc(*fourcc)
    writer = cv2.VideoWriter(str(path), fourcc, fps, size.to_tuple_int())

    dst_img = ImageNda(size)
    for i, src_img in enumerate(images):
        resize(src_img, dst_img)
        # print('write', i, dst_img.data().shape)
        writer.write(dst_img.data())

    writer.release()


def save_images(
    images: ImageNdas, path: StrPath, fps: float, size: Size = Size()
) -> None:
    """保存图片集合成动画/视频文件"""
    assert images, "没有待保存的图片"
    path = Path(path)
    # print(path.suffix)
    match path.suffix:
        case ".mp4":
            fun = _save_images_iio
        case ".gif" | ".pdf" | ".png" | ".tiff" | ".webp":
            fun = _save_images_pil
        case _:
            raise RuntimeError("不支持的文件格式")
    make_parents(path)

    size = size or images[0].size()
    images = [resize(im, dst_size=size) for im in images]
    fun(images, path, fps)


def _save_images_iio(images: ImageNdas, path: Path, fps: float) -> None:
    """保存图片集合成动画/视频文件"""

    fourcc_tab = {".mp4": "h264"}  # require libwebp-dev
    codec = fourcc_tab.get(path.suffix)
    assert codec, "不支持的文件格式"

    writer = iio.get_writer(str(path), format="FFMPEG", mode="I", fps=fps, codec=codec)  # type: ignore

    for im in images:
        writer.append_data(im.data())

    writer.close()


def _save_images_pil(images: ImageNdas, path: Path, fps: float) -> None:
    """保存图片集合成动画/视频文件"""
    ims = [bgr_to_pil(im.data()) for im in images]
    # print(ims)
    duration = int(1 / fps * 1000)
    ims[0].save(path, save_all=True, append_images=ims[1:], duration=duration, loop=0)


def load_images_in(folder: StrPath, ext: str = ".jpg") -> ImageNdas:
    """加载目录下所有指定图片文件"""
    files = files_in(folder, ext)
    return [ImageNda.load(f) for f in files]


def load_image_pairs_in(
    folder: StrPath, ext: str = ".jpg"
) -> list[tuple[Path, ImageNda]]:
    """加载目录下所有指定图片文件, 返回路径/图片对数组"""
    files = files_in(folder, ext)
    return [(f, ImageNda.load(f)) for f in files]


class Capture:
    """推向捕获器"""

    def __init__(self, path: StrPath):
        self._capture = cv2.VideoCapture(str(path))

    def at(self, seconds: float) -> Optional[ImageNda]:
        """获取指定位置(时间)的图像"""
        ms = int(seconds * 1000)
        self._capture.set(cv2.CAP_PROP_POS_MSEC, ms)
        ok, frame = self._capture.read()
        return ImageNda(data=frame) if ok else None

    def position(self) -> float:
        """获取位置, 单位秒"""
        return self._capture.get(cv2.CAP_PROP_POS_MSEC) / 1000



#!/opt/ias/env/bin/python

import argparse
from pathlib import Path

import arrow
import cv2
from arrow import Arrow
from numpy import ndarray

from jcx.sys.fs import time_to_file, make_parents
from jcx.ui.key import Key
from jvi.geo.point2d import Point
from jvi.geo.size2d import size_parse, Size
from jvi.gui.video_viewer import VideoViewer

TITLE = "video2image"


class ImageCapture(VideoViewer):
    """图片抓取"""

    def __init__(
        self, url, dst_dir: Path, size: Size = Size(), start_time: Arrow = arrow.now()
    ):
        super().__init__(url, TITLE, size)
        self.__dst_dir = dst_dir
        self.__start_time = start_time

    def on_key(self, key: int):
        """按键响应"""
        if key == Key.BLANK:
            self.__save_image()
        else:
            return super().on_key(key)
        return 0

    def on_draw(self, canvas: ndarray, pos: Point):
        """设置窗口重绘事件响应"""
        # self.record().draw_on(canvas, pos)
        pass

    def __save_image(self):
        """保存图片"""
        image = self.image_at(self._index)
        if image is None:
            print("[ERROR] image is none")
            return

        interval = self._index * 1000000 / 25
        time = self.__start_time.shift(microseconds=interval)
        file = self.__dst_dir / time_to_file(time, ".jpg")
        make_parents(file)
        cv2.imwrite(str(file), image)
        print("save to:", file)


# file = 'file:///home/jiang/ws/lift/2021-08-19/1.mp4'


def main():
    parser = argparse.ArgumentParser(description="视频转图片工具")
    parser.add_argument("src_url", type=Path, help="来源视频URL")
    parser.add_argument("dst_dir", type=Path, help="目的图片所在目录")
    parser.add_argument("-e", "--ext", type=str, default=".jpg", help="图片文件扩展名")
    parser.add_argument("-s", "--size", type=str, default="HD", help="输出视频尺寸(HD)")
    parser.add_argument("-i", "--interval", type=float, default=1, help="时间间隔")
    parser.add_argument("-t", "--start_time", type=str, help="图片开始时间")
    parser.add_argument("-v", "--verbose", action="store_true", help="显示详细信息")
    opt = parser.parse_args()

    size = size_parse(opt.size)
    # TODO: start_time解析
    win = ImageCapture(opt.src_url, opt.dst_dir, size=Size(1280, 720))
    if not win.is_opened():
        print("[ERROR] cannot open url:", opt.src_url)
    print("frame count:", win.image_count())
    print("video size:", win.image_size())
    # print(opt.start_time)

    win.run()


if __name__ == "__main__":
    main()



#!/opt/ias/env/bin/python

import argparse
from pathlib import Path

import cv2
import numpy as np

from jcx.ui.key import Key
from jml.iqa.a import estimate_clearness


def trans(src):
    gray = cv2.cvtColor(src, cv2.COLOR_BGR2GRAY)
    print(gray.dtype)
    edge = cv2.Sobel(gray, -1, 1, 1, ksize=3)
    print("edge:", edge.shape, edge.dtype)
    hist = cv2.calcHist([edge], [0], None, [256], [0, 256])
    # edge = cv2.Canny(gray, 100, 200)
    hist *= 255 / np.sum(hist)
    hist = hist.astype(np.uint8)
    hist = hist[:224]
    print("hist:", hist.shape, np.sum(hist))

    print("t:", np.transpose(hist))

    hist = np.repeat(hist, 224, axis=1)
    print("hist:", hist.shape, np.sum(hist))

    return hist


def main():
    parser = argparse.ArgumentParser(description="图片模糊度评估工具")
    parser.add_argument("src_dir", type=Path, help="来源图片目录")
    parser.add_argument("dst_dir", type=Path, help="目的图片目录")
    parser.add_argument(
        "-p", "--pattern", type=str, default="*.jpg", help="图片文件模式(*.jpg)"
    )
    parser.add_argument("-t", "--tail", type=str, default="_b", help="图片后缀")
    parser.add_argument("-v", "--verbose", action="store_true", help="显示详细信息")
    opt = parser.parse_args()

    files = opt.src_dir.rglob(opt.pattern)
    print("源图片目录：%s" % opt.src_dir)

    colors = []

    for i, f in enumerate(files):
        src = ImageNda.load(str(f))  # BGR
        if src is None:
            print(i, f.name, "\tERROR")
            continue
        if opt.verbose:
            print(i, f.name, "\tOK")

        color = estimate_clearness(src, 0.9)
        colors.append(color)
        # print('#%d' % i, color)

        if opt.verbose and color == 0:
            cv2.imshow("src", src)
            if cv2.waitKey(0) == Key.ESC:
                break

    print("range: [%d, %d] mean: %d" % (min(colors), max(colors), np.mean(colors)))
    cv2.destroyAllWindows()


if __name__ == "__main__":
    main()
